#include <amscimglib4/amscimglib4.h>
#include <amscimglib4/amscimglib4_intl.h>

#ifdef __cplusplus
extern "C" {  
#endif

void amscimglib4_test_init()
{
    amscimglib4_image *img1 = NULL;
    amscimglib4_image *img2 = NULL;
    int res;

    printf("Testing basic image operations for memory leaks...\n");
    printf("use valgrind...\n");
    
    res = amscimglib4_image_new(&img1,1000,1000); printf("res=%d (%d,%d)\n",res,img1->sizex,img1->sizey);
    res = amscimglib4_image_new(&img2,200,1000); printf("res=%d (%d,%d)\n",res,img2->sizex,img2->sizey);
    
    res = amscimglib4_image_resize(img1,500,500); printf("res=%d (%d,%d)\n",res,img1->sizex,img1->sizey);
    res = amscimglib4_image_copy(img2,img1); printf("res=%d, img2=(%d,%d), img1=(%d,%d)\n",res, img2->sizex,img2->sizey,img1->sizex,img1->sizey);
    res = amscimglib4_image_resize(img1,-10,3); printf("res=%d (%d,%d)\n",res,img1->sizex,img1->sizey);
    res = amscimglib4_image_resize(img1,0,0); printf("res=%d (%d,%d)\n",res,img1->sizex,img1->sizey);
    res = amscimglib4_image_copy(img2,img1); printf("res=%d, img2=(%d,%d), img1=(%d,%d)\n",res, img2->sizex,img2->sizey,img1->sizex,img1->sizey);
    


    amscimglib4_image_delete(&img1);
    amscimglib4_image_delete(&img2);
    
}

AMSCIMGLIB4_API void amscimglib4_test_loadsave()
{
    printf("Testing basic image loading/saving and conversion...\n");

    amscimglib4_image *img1 = NULL;
    amscimglib4_image *img2 = NULL;

    amscimglib4_image_new(&img1,1,1);
    amscimglib4_image_new(&img2,1,1);

    amscimglib4_readimage_bmp("../testimg/testpfp0.bmp",img1);
    printf("read in (%d,%d) bitmap image.\n",img1->sizex,img1->sizey);
    amscimglib4_image_flipx(img1);
    amscimglib4_writeimage_bmp("../testimg/testpfp0_1.bmp",img1);
    amscimglib4_writeimage_jpeg("../testimg/testpfp0_2.jpg",img1,90);
    amscimglib4_writeimage_png("../testimg/testpfp0_3.png",img1);

    amscimglib4_readimage("../testimg/testpfp0.png",img1);
    printf("read in (%d,%d) png image.\n",img1->sizex,img1->sizey);
    amscimglib4_transpose_image(img1);
    amscimglib4_writeimage("../testimg/testpfp1_1.bmp",img1);
    amscimglib4_writeimage("../testimg/testpfp1_2.png",img1);
    amscimglib4_writeimage("../testimg/testpfp1_3.jpg",img1);

    amscimglib4_readimage("../testimg/testpfp0.jpg",img1);
    printf("read in (%d,%d) jpg image.\n",img1->sizex,img1->sizey);
    amscimglib4_image_flipy(img1);
    amscimglib4_writeimage("../testimg/testpfp2_1.bmp",img1);
    amscimglib4_writeimage("../testimg/testpfp2_2.png",img1);
    amscimglib4_writeimage("../testimg/testpfp2_3.jpg",img1);

    amscimglib4_image_delete(&img1);
    amscimglib4_image_delete(&img2);

}

AMSCIMGLIB4_API void amscimglib4_test_binreadwrite()
{
    int res;
    unsigned char *bytebuff = NULL;
    long bufflen = 10000000;
    long nread = 0;
    const char *fname1 = "../testimg/testpfp0.png";
    const char *fname2 = "../testimg/testpfp0_bwrite.png";
    const char *fname3 = "../testimg/testpfp0_bwrite2.png";
    bytebuff = (unsigned char *) malloc(sizeof(unsigned char)*bufflen);
    if(bytebuff==NULL)
    {
        printf("bytebuff failed to allocate.\n");
        return;
    }

    amscimglib4_image *img1 = NULL;
    amscimglib4_image *img2 = NULL;

    amscimglib4_readbinfile(fname1,bytebuff,bufflen,&nread);
    printf("Read %ld bytes into byte buffer from %s\n",(long)nread,fname1);

    amscimglib4_writebinfile(fname2,bytebuff,nread);

    amscimglib4_image_new(&img1,1,1);
    amscimglib4_image_new(&img2,1,1);


    //every so often this segfaults for some reason
    res = amscimglib4_pngbytebuffer_to_img(img1,bytebuff,nread);
    printf("check something: %d\n",(int)sizeof(unsigned char));

    printf("Interpreted png as (%d, %d) image.\n",(int)img1->sizex,(int)img1->sizey);

    memset(bytebuff,0,bufflen);
    nread = amscimglib4_pngbytebufferlen(img1);
    printf("To write the image to a bytebuffer, I need %ld bytes\n",nread);

    res = amscimglib4_pngbytebuffer_from_img(img1,bytebuff,bufflen,&nread);
    printf("Wrote %ld bytes to buffer\n.",nread);

    amscimglib4_writebinfile(fname3,bytebuff,nread);

    amscimglib4_image_delete(&img1);
    amscimglib4_image_delete(&img2);
    
    free(bytebuff);

    return;
}

#ifdef __cplusplus
}; //end extern "C"
#endif